@@ -1,6 +1,5 @@
 from collections import defaultdict
 from pyclbr import Function
-from random import random
 from tokenize import Double
 from typing import Dict
 from xmlrpc.client import Boolean
@@ -55,7 +54,7 @@ class MVNMixtureModel():
             # mean and sd for the Normal prior of the variance
             "var_loc":torch.tensor(55).float(), \
             "var_scale":torch.tensor(30).float(), \
-            "min_var":torch.tensor(20).float(),
+            "min_var":torch.tensor(5).float(),
             "eta":torch.tensor(1).float()}
         self._autoguide = False
         self._enumer = "parallel"
@@ -145,7 +144,7 @@ class MVNMixtureModel():
         self.lm["slope"], self.lm["intercept"] = slope, intercept
 
 
-    def _filter_dataframe_init(self, min_ccf=.05, k_interval=(5,25), \
+    def _filter_dataframe_init(self, min_ccf=.05, k_interval=(5,15), \
             metric="calinski_harabasz_score", random_state=25):
         '''
         Function to filter the input dataset according to the centroid the clusters output
@@ -348,7 +347,7 @@ class MVNMixtureModel():
 
         return var, var_constr
 
-    
+
     def _initialize_sigma_chol(self, K):
         if self.cov_type == "diag" or self._T == 1:
             sigma_chol = torch.eye(self._T) * 1.
@@ -388,11 +387,13 @@ class MVNMixtureModel():
         return self.init_params
 
 
-    def fit(self, steps=500, optim_fn=pyro.optim.Adam, lr=0.001, cov_type="diag", \
+    def fit(self, steps=500, optim_fn=pyro.optim.Adam, lr=0.005, cov_type="diag", \
             loss_fn=pyro.infer.TraceEnum_ELBO(), convergence=True, initializ=True, \
-            min_steps=1, p=0.01, random_state=25, store_params=False, show_progr=True):
+            min_steps=1, p=.5, random_state=25, store_params=False, show_progr=True):
+        
         pyro.enable_validation(True)
-        pyro.clear_param_store()
+        # pyro.clear_param_store()
+        pyro.get_param_store().__init__()
 
         self._settings = {"optim":optim_fn({"lr":lr, "betas": (0.93, 0.999)}), "loss":loss_fn, "lr":lr}
         self._is_trained = False
@@ -438,8 +439,7 @@ class MVNMixtureModel():
         return self.svi.loss(self.model, self._global_guide)
 
 
-    def _train(self, steps, convergence, min_steps=1, p=0.01, \
-        show_progr=True, store_params=False):
+    def _train(self, steps, convergence, min_steps, p, show_progr=True, store_params=False):
         '''
         Function to perform the training of the model. \\
         `steps` is the maximum number of steps to be performed. \\
@@ -471,7 +471,7 @@ class MVNMixtureModel():
                 for k,v in params_step.items():
                     if k in ["weights", "mean", "sigma_vector", "sigma_chol"]:
                         params[k] = params.get(k, dict())
-                        params[k]["step_"+str(step)] = v.detach().numpy()
+                        params[k]["step_"+str(step)] = v.clone().detach().numpy()
 
             # gradient_norms = self._reset_params(params=params_step, p=.01, gradient_norms=gradient_norms)
             if convergence and step >= min_steps:
@@ -479,7 +479,7 @@ class MVNMixtureModel():
                 sigma_conv[0], sigma_conv[1] = sigma_conv[1], params_step["sigma_vector"]
                 conv = self._convergence(mean_conv, sigma_conv, conv, p=p)
                 # conv = self._convergence_grads(gradient_norms, conv)
-                if conv == 10:
+                if conv == 5:
                     if show_progr:
                         t.set_description("ELBO %f" % elb)
                         t.reset(total=step)
@@ -494,15 +494,15 @@ class MVNMixtureModel():
             "params":params}
 
 
-    def _convergence_grads(self, gradient_norms, conv, p=0.005):
-        for gr in gradient_norms.keys():
-            if gr in ["mean_param", "sigma_vector_param"]:
-                prev = gradient_norms[gr][-2]
-                curr = gradient_norms[gr][-1]
-                cc = np.abs(prev - curr) < curr*p
-        if cc:
-            return conv +1
-        return 0
+    # def _convergence_grads(self, gradient_norms, conv):
+    #     for gr in gradient_norms.keys():
+    #         if gr in ["mean_param", "sigma_vector_param"]:
+    #             prev = gradient_norms[gr][-2]
+    #             curr = gradient_norms[gr][-1]
+    #             cc = np.abs(prev - curr) < curr*p
+    #     if cc:
+    #         return conv +1
+    #     return 0
 
 
     def _convergence(self, mean_conv, sigma_conv, conv, p):
@@ -511,7 +511,17 @@ class MVNMixtureModel():
         return 0
 
 
-    def _check_convergence(self, par, p=0.01) -> Boolean:
+    def _normalize(self, par):
+        # zi = (xi - min(x)) / (max(x) - min(x))
+        norm = list()
+        for p in par:
+            if torch.min(p) == torch.max(p):
+                norm.append(p)
+            norm.append(( p - torch.min(p) ) / ( torch.max(p) - torch.min(p) ) * 100)
+        return norm
+
+
+    def _check_convergence(self, par, p) -> Boolean:
         '''
         - `par` -> list of 2 elements given the previous (at index 0) and current (at index 1) 
         estimated values for a parameter.
@@ -519,14 +529,19 @@ class MVNMixtureModel():
         The function returns `True` if more than 95% of the values changed less than `perc*100`% 
         in the current step with respect to the previous one.
         '''
+        par = self._normalize(par)
         eps = p * self._settings["lr"]
+        # print(eps)
         n = 0
         for k in range(par[0].shape[0]):
             for t in range(par[0].shape[1]):
-                perc = eps * torch.max(torch.tensor(1), par[0][k,t])
+                perc = eps * par[0][k,t] # torch.max(torch.tensor(1), par[0][k,t])
                 if torch.absolute(par[0][k,t] - par[1][k,t]) <= perc:
                     n += 1
-        return n >= .99 * self.params["K"]*self.params["T"]
+                # else:
+                #     print(par[0][k,t], par[1][k,t], par[0][k,t] - par[1][k,t], eps * par[0][k,t])
+
+        return n >= .9 * self.params["K"]*self.params["T"]
 
 
     def _get_learned_parameters(self) -> Dict:
@@ -541,11 +556,11 @@ class MVNMixtureModel():
 
         p = {}
         p["N"], p["K"], p["T"] = self.params["N"], self.params["K"], self.params["T"]
-        p["weights"] = param_store["weights_param"]
+        p["weights"] = param_store["weights_param"].clone().detach()
 
-        p["mean"] = param_store["mean_param"]
-        p["sigma_vector"] = param_store["sigma_vector_param"]
-        p["sigma_chol"] = param_store["sigma_chol_param"]
+        p["mean"] = param_store["mean_param"].clone().detach()
+        p["sigma_vector"] = param_store["sigma_vector_param"].clone().detach()
+        p["sigma_chol"] = param_store["sigma_chol_param"].clone().detach()
 
         if self._is_trained:
             p["sigma"] = self.compute_Sigma(p["sigma_chol"], p["sigma_vector"], p["K"])
